/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.blackberry;

import java.util.Vector;
import java.util.Enumeration;

import net.rim.device.api.system.RadioInfo;

import com.funambol.blackberry.BBSyncManager;
import com.funambol.blackberry.controller.SynchronizationController;
import com.funambol.blackberry.source.Poller;
import com.funambol.blackberry.AppSyncSource;
import com.funambol.syncml.protocol.SyncML;
import com.funambol.syncml.spds.CompressedSyncException;
import com.funambol.syncml.spds.SyncConfig;
import com.funambol.syncml.spds.SyncException;
import com.funambol.syncml.spds.SyncManager;
import com.funambol.syncml.spds.SyncSource;
import com.funambol.blackberry.source.BBPIMSyncSource;
import com.funambol.syncml.protocol.SyncMLStatus;
import com.funambol.util.BasicConnectionListener;
import com.funambol.util.Log;
import com.funambol.util.ConnectionConfig;
import com.funambol.util.ConnectionManager;
import com.funambol.util.BlackberryUtils;
import com.funambol.util.StringUtil;

public class SyncEngine {

    private int                             scheduledAttempt = 0;

    private Poller                          retryPoller;

    private final SynchronizationController controller;

    public SyncEngine(SynchronizationController controller) {
        this.controller = controller;
    }

    public boolean synchronize(Vector sources) {

        Log.info("[SyncEngine.synchronize]");

        controller.clearErrors();

        if (ConnectionConfig.isUserConfirmationNeeded()) {
            Log.debug("Setting connection listener for this application");
            ConnectionManager.getInstance().setConnectionListener(controller);
        } else {
            Log.debug("Using Default BasicConnectionListener");
        }

        if (   Configuration.username == null || Configuration.username.length() == 0
            || Configuration.password == null || Configuration.password.length() == 0) {

            controller.noCredentials();
            return false;
        }

        controller.setCancel(false);

        Log.info("SyncClient.synchronize(int): Entering - synchronize flag is "
                + controller.isSynchronizing());

        // Safety check. If there are no ready to use sources the sync is
        // stopped. This case should be captured earlier in the flow, but
        // just in case....
        if (AppSyncSourceManager.getInstance().numberOfEnabledAndWorkingSources() == 0) {
            controller.noSources();
            return false;
        }

        boolean signal = BlackberryUtils.isDataConnectionAvailable();

        if (!signal) {
            if (RadioInfo.getState() == RadioInfo.STATE_OFF) {
                controller.noConnection();
            } else {
                controller.noSignal();
            }
            return false;
        }

        Vector checkSources = new Vector();

        for (int x = 0; x < sources.size(); x++) {
            AppSyncSource appSource = (AppSyncSource) sources.elementAt(x);
            SyncSource    source    = appSource.getSyncSource();
            int mode = source.getConfig().getSyncMode();
            if (   mode != SyncML.ALERT_CODE_REFRESH_FROM_CLIENT
                && mode != SyncML.ALERT_CODE_REFRESH_FROM_SERVER) {
                checkSources.addElement(source);
            }
        }

        if (Customization.getInstance().warnOnDeletes()) {
            Vector sourcesToRemove = checkDeletes(checkSources);
            for (int x = 0; x < sourcesToRemove.size(); x++) {
                sources.removeElement(sourcesToRemove.elementAt(x));
                x--;
            }
        }

        if (sources.size() == 0) {
            controller.noSources();
            return false;
        } else {
            if (controller.isCancelled()) {
                return false;
            }

            final SyncThread syncThread = new SyncThread(sources);
            syncThread.setPriority(Thread.MIN_PRIORITY);

            syncThread.start();
        }// end if

        return true;
    }

    /**
     * Called to warn the user of a large number of deletes
     * 
     * @param sources list of AppSyncSources for which we shall warn the user
     * @returns Whether or not to continue with the sync
     */
    private boolean warnDelete(Enumeration sources) {
        boolean result = controller.confirmDeletes(sources);
        return result;
    }

    /**
     * Check to see if the user should be warned, and if so, warn them.
     * 
     * @param mask
     */
    private Vector checkDeletes(Vector sources) {

        boolean toWarn = false;
        Vector warnSources = new Vector();

        for (int x = 0; x < sources.size(); x++) {
            AppSyncSource appSource = (AppSyncSource)sources.elementAt(x);
            SyncSource source = appSource.getSyncSource();
            if (source instanceof BBPIMSyncSource) {
                BBPIMSyncSource pimSyncSource = (BBPIMSyncSource)source;
                if (pimSyncSource.shouldWarn()) {
                    toWarn = true;
                    warnSources.addElement(appSource);
                }
            }
        }

        if (toWarn) {
            if (!warnDelete(warnSources.elements())) {
                return sources;
            }
        }

        return new Vector();
    }

    private class SyncThread extends Thread {

        private final  Vector appSources;
        boolean               compressionRetry;

        private SyncConfig   syncConfig;

        public SyncThread(Vector sources) {
            this.appSources       = sources;
            this.compressionRetry = false;
        }

        public synchronized void run() {

            Log.info("[SyncThread.run]");

            syncConfig = controller.getSyncConfig();
            if (compressionRetry) {
                // We are trying without compression, make sure compression is
                // really disabled
                syncConfig.compress = false;
            }
            controller.syncStarted(appSources);
            Vector failedSources = new Vector();

            try {

                Log.info("SyncClient.synchronize(int): Starting sync thread - synchronize flag is "
                        + controller.isSynchronizing());
                failedSources = synchronize();

            } catch (CompressedSyncException e) {

                if (!compressionRetry) {
                    // Only retry because of compression once
                    Log.info("Sync failed because compression failed - Retrying");
                    compressionRetry = true;

                    // Recurse
                    this.run();
                    return;
                } else {
                    controller.error("Sync failed with compression error after disabling compression");
                }
            } catch (Exception e) {
                // This is unexpected, but we don't want the app to die
                // The finally block will take care of updating the status of
                // the failed sources. Since this case is unexpected we do not
                // rely on the failedSource returned value but we signal all the
                // sources as failed
                failedSources = appSources;
                Log.error("Exception caught during synchronization " + e.toString());
            } finally {
                controller.syncEnded(appSources, null);
            }
        }

        /**
         * The main procedure for the sync thread
         * 
         * @param syncMask
         * @throws Exception
         */
        private Vector synchronize() throws Exception {

            Log.info("[SyncEngine.synchronize]");

            // Uncomment this line to enable automatic wifi usage at every sync
            // ConnectionConfig.removeSavedConfig();
            Log.info(ConnectionConfig.getConfigsDescription());
            Log.info(ConnectionConfig.getWorkingConfigurationDescription());

            BBSyncManager manager = new BBSyncManager(syncConfig);
            Vector failedSources = new Vector();

            for (int x = 0; x < appSources.size(); x++) {


                if (controller.isCancelled()) {
                    // If the sync got cancelled then we must exit the loop
                    // Even if the user cancel the sync the SyncSource may be
                    // unable to throw the proper exception if other errors
                    // (such as network error) kick in. So we must recheck here
                    // and stop the sync if necessary
                    break;
                }

                AppSyncSource appSource = (AppSyncSource)appSources.elementAt(x);
                SyncSource source = appSource.getSyncSource();

                controller.sourceStarted(appSource);

                try {

                    // TODO not always necessary
                    manager.setFlagSendDevInf();
                    manager.sync(source, source.getConfig().getSyncMode());

                    if (source.getStatus() != SyncSource.STATUS_SUCCESS) {
                        failedSources.addElement(appSource);
                    }

                    // If we get here, it means that the sync was succesfull and
                    // we can update the anchors
                    controller.sourceEnded(appSource);
                } catch (final Exception e) {

                    if (source instanceof BBPIMSyncSource) {
                        BBPIMSyncSource pimSource = (BBPIMSyncSource)source;
                        pimSource.restoreTrackingData();
                    }

                    boolean compressError = (e instanceof CompressedSyncException);
                    if (compressError) {
                        if (!compressionRetry) {
                            Log.info("Retrying without compressio");
                        }

                        throw new CompressedSyncException(e.getMessage());
                    }

                    SyncException se;
                    if (e instanceof SyncException) {
                        se = (SyncException)e;
                    } else {
                        se = new SyncException(SyncException.CLIENT_ERROR, e.toString());
                    }

                    controller.sourceFailed(appSource, se);

                    // After this point, we know the source really
                    // failed

                    failedSources.addElement(appSource);

                    // Depending on the reason why the sync failed, we may
                    // want to abort the other sources as well (e.g. invalid
                    // credentials)
                    if (se.getCode() == SyncException.AUTH_ERROR ||
                        se.getCode() == SyncException.FORBIDDEN_ERROR ||
                        se.getCode() == SyncException.CANCELLED) {

                        break;
                    }
                } finally {
                    source = null;
                }
            }

            controller.setCancel(false);

            // -- END OF A SYNCHRONIZATION
            controller.syncComplete(appSources, failedSources.size() > 0);

            return failedSources;
        }

    }

    public boolean retry() {
        boolean willRetry = false;

        if (retryPoller != null) {
            retryPoller.disable();
        }

        if (scheduledAttempt < 3) {
            scheduledAttempt++;
            Log.error("Scheduled sync: Connection attempt failed. " + "Try again between "
                    + Constants.RETRY_POLL_TIME + " minutes");

            retryPoller = new Poller(controller, Constants.RETRY_POLL_TIME, true);
            retryPoller.start();
            willRetry = true;
        } else {
            retryPoller.disable();
            retryPoller = null;
            scheduledAttempt = 0;
        }
        return willRetry;
    };
}
