/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.blackberry.source;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import javax.microedition.pim.PIMException;
import javax.microedition.pim.PIMItem;
import javax.microedition.pim.PIMList;
import javax.microedition.pim.Contact;

import net.rim.device.api.system.ControlledAccessException;
import net.rim.device.api.system.PersistentObject;
import net.rim.device.api.system.PersistentStore;
import net.rim.device.api.util.Persistable;

import com.funambol.blackberry.controller.SynchronizationController;
import com.funambol.blackberry.helper.PIMItemHelper;
import com.funambol.syncml.protocol.SyncMLStatus;
import com.funambol.syncml.spds.SourceConfig;
import com.funambol.syncml.spds.SyncException;
import com.funambol.syncml.spds.SyncItem;
import com.funambol.syncml.protocol.SyncML;
import com.funambol.util.Log;

/**
 * Base class for PIM sync source in BB. This class is a specialization of
 * BBPIMSyncSource with a workaround for a bug in RIM's APIs on OS 4.5. The
 * class can be used on any platform, but its usage is discourages unless
 * necessary.
 * The main feature provided is the ability to have different values for LUID
 * and UID. The UID is assigned by the system to any created PIM Item, the LUID
 * is the identifier of an item that is sent to the server by the Sync Engine.
 * The best choice is to have UID=LUID and this is what the standard
 * BBPIMSyncSource does. But on OS 4.5 there is a problem in the APIs that
 * sometimes are unable to clean a contact. Clearing its fields may generate
 * an unexpcted IndexOutOfBoundsException and then updates from the server
 * cannot be safely applied (in other cases there is no exception but the
 * contact is not cleared completely). To workaround this problem, this class
 * has a mapping between luid and uid and updates are implemented by removing
 * the contact and recreating it. Since the newly created contact has a
 * different UID, we need a map to know its LUID to handle operations with the
 * server.
 * At the interface level this class works with LUID. So any item that comes in
 * or goes out has a key equal to the LUID. Internally the class maps LUID to
 * UID (and viceversa) to communicate with "helpers" which operate on UIDs.
 */
public class BBPIMSyncSourceWithUpdateFix extends BBPIMSyncSource
{
    private int syncMode;

    private Hashtable preloadedItems = null;

    public BBPIMSyncSourceWithUpdateFix(SourceConfig config,
                                        PIMList list, PIMItemHelper helper,
                                        long persistenceStorageKey,
                                        long addAttributesPersistenceStorageKey) throws SyncException
    {
        super(config, list, helper, persistenceStorageKey,addAttributesPersistenceStorageKey);
    }


    //----------------------------------------------------------- Public Methods

    public void beginSync(int syncMode) throws SyncException {
        this.syncMode = syncMode;
        super.beginSync(syncMode);
    }

    /*
     * (non-Javadoc)
     * @see com.funambol.syncml.spds.SyncSource#updateItem(com.funambol.syncml.spds.SyncItem)
     */
    public int updateItem(SyncItem item) throws SyncException
    {
        Log.info("[updateItem] " + item.getKey());
        totalUpdated++;
        
        if (controller.isCancelled())  cancelSync();
        
        try
        {
            String luid = item.getKey();
            String uid = getUIDFromLuid(luid);
            PIMItem pitem = findItem(uid);
            AdditionalItemsAttributes additionals = null;
            if(additionalsMap != null){
                additionals = (AdditionalItemsAttributes) additionalsMap.get(uid);
            }
            if (pitem!=null)
            {
                // We delete the item, create an empty one and add a mapping
                // Get the UID from the luid
                // Before removing the item we must save the fields that
                // are not synchronized so that we can restore them in the newly
                // created item
                PIMItem origItem = pitem;

                pitem = helper.createItem();
                AdditionalItemsAttributes newAdditionals = helper.fillItem(pitem,new String(item.getContent()),additionals);

                // During a slow sync most updates are due to the fact that the
                // BB supports only a subset of the fields of a contact.
                // Therefore the server sends the items as updated, but then the
                // BB drops these extra fields again. Since removing items on OS
                // 4.5 is very costly, we check first if the received item is
                // identical to the one we already have. In such a case, we
                // simply return a success code and ignore the update.
                // This kind of twin detection is performed by parsing and
                // formatting back the incoming item.
                if (syncMode == SyncML.ALERT_CODE_SLOW) {
                    // If this is a contact and it has a picture, then it is faster to
                    // delete/add it, rather than formatting it twice
                    if (origItem instanceof Contact && origItem.countValues(Contact.PHOTO) > 0) {
                        Log.info("Skip items comparison");
                    } else {
                        String origItemText = helper.itemToString(origItem, additionals);
                        String newItemText = helper.itemToString(pitem, newAdditionals);

                        if (newItemText.equals(origItemText)) {
                            Log.info("The new item is identical to the old one, skipping the update");
                            return SyncMLStatus.SUCCESS;
                        } else {
                            Log.info("Slow sync has a real update");
                        }
                    }
                }

                // Copy all the fields that are not synchronized otherwise they
                // get lost (this must be done before removing the item from the
                // list)
                try {
                    helper.copyUnsupportedFields(origItem, pitem);
                } catch (Exception e) {
                    Log.error("Error while preserving unsupported fields: " + e.toString());
                    // We ignore this error and continue the synchronization
                }

                Log.debug("Deleting item");
                helper.deleteItem(origItem);
                removeAdditionals(uid);

                trackingData.deleteDeletedKey(uid,true);
                trackingDataBackup.deleteDeletedKey(uid);
      
                Log.trace("Before commiting PIMItem.");
                try {
                    helper.saveUpdatedPIMItem(pitem);
                    setAdditionals(helper.getUID(pitem), newAdditionals);
                } catch (PIMException e) {
                    String error = helper.handlePIMException(e);
                    logError("Unable to save updated item in source [" + getName() + "]: " + error);
                    return SyncMLStatus.GENERIC_ERROR;
                }
                // These operations must be done before changing the 
                // LUID
                trackingData.deleteNewKey(helper.getUID(pitem),true);
                trackingDataBackup.deleteNewKey(helper.getUID(pitem));
                

                item.setKey(luid);
                return SyncMLStatus.SUCCESS;
            }
            logError("Unable to save updated item in source [" + getName() + "]: item not found");
            return SyncMLStatus.NOT_FOUND;
        } catch(final PIMException e) {
            String error = helper.handlePIMException(e);
            logError("PIMException while updating item to SyncSource [" + getName() + "]: " + error);
            return SyncMLStatus.GENERIC_ERROR;
        } catch (Exception e) {
            // We had some sort of generic runtime exception
            logError("RuntimeException while updating item to SyncSource ["
                      + getName() + "]: " + e.toString());
            return SyncMLStatus.GENERIC_ERROR;
        }
    }
    
    /*
     * (non-Javadoc)
     * @see com.funambol.syncml.spds.SyncSource#deleteItem(java.lang.String)
     */
    public int deleteItem(String item) throws SyncException
    {
        Log.info("[deleteItem] " + item);
        totalDeleted++;

        if (controller.isCancelled())  cancelSync();
        
        try
        {
            String luid = item;
            String uid = getUIDFromLuid(luid);
            Log.debug("Before trying to find and delete existing PIMItem.");
            if (helper.deleteItemFromKey(uid)!=null)
            {
                Log.trace("Purposfully deleting deleted item key: " + uid);
                trackingData.deleteDeletedKey(uid, true);
                trackingDataBackup.deleteDeletedKey(uid);

                removeAdditionals(uid);
                
                return SyncMLStatus.SUCCESS;
            }
            logError("Unable to delete item in source [" + getName() + "]: no error");
            return SyncMLStatus.NOT_FOUND;
        }
        catch(final Exception e)
        {
            logError("Exception while deleting item from SyncSource ["+getName()+"]");
            return SyncMLStatus.GENERIC_ERROR;
        }
    }


    /**
     * This method is invoked by the BBSyncManager once the modifications from
     * the server have been preprocessed and we know that items are going to be
     * updated/deleted. These items are preloaded here with a linear scan over
     * the list of PIM items.
     */
    public void prepareModifications(Vector itemsKey) {
        Log.info("[prepareModifications]");

        preloadedItems = new Hashtable();
        try {
            Enumeration items = helper.getItemsList();
            // Perform a single scan of all the items and load the ones that
            // are going to be updated/deleted
            while(items.hasMoreElements())
            {
                PIMItem item = (PIMItem)items.nextElement();
                String itemUid = helper.getUID(item);
                // Check if this item is part of the one to be updated/changed
                for(int i=0;i<itemsKey.size();++i) {
                    String luid = (String) itemsKey.elementAt(i);
                    String uid  = getUIDFromLuid(luid);

                    if (itemUid != null && itemUid.equals(uid)) {
                        // This item must be pre-loaded
                        Log.info("[BBPIMSyncSourceWithUpdateFix.prepareModifications] preloading: " + uid);
                        preloadedItems.put(uid, item);
                    }
                }
            }
        } catch (Exception e) {
            Log.error("Error while preloading items: " + e.toString());
        }
    }

    /*
     * (non-Javadoc)
     * @see com.funambol.syncml.spds.SyncSource#getItemContent(com.funambol.syncml.spds.SyncItem)
     */
    protected SyncItem getItemContent(SyncItem item) throws SyncException
    {
        Log.info("[getItemContent] " + item.getKey());
        
        if (controller.isCancelled())  cancelSync();
        
        try
        {
            String luid = item.getKey();
            String uid = getUIDFromLuid(luid);
            Log.debug("luid=" + luid + " uid = " + uid);
            PIMItem pitem = helper.findItem(uid);
            AdditionalItemsAttributes additionals = null;
            if(additionalsMap != null){
                additionals = (AdditionalItemsAttributes) additionalsMap.get(uid);
            }
            if (pitem!=null)
            {
                item.setContent(helper.itemToString(pitem, additionals).getBytes());
                return item;
            }
        }
        catch(final PIMException e)
        {
            throw new SyncException(SyncException.CLIENT_ERROR,
                                    "Exception while retrieving content of item from SyncSource ["
                                    +getName()+"]");
        }
        return null;
    }

   
   

    private String getUIDFromLuid(String luid) throws PIMException, SyncException {

        String uid = null;
        // We must do a reverse search in the map
        Enumeration keys = additionalsMap.keys();
        while(keys.hasMoreElements()) {
            String key = (String)keys.nextElement();
            AdditionalItemsAttributes addits   = (AdditionalItemsAttributes)additionalsMap.get(key);
            if (luid.equals(addits.getLuid())) {
                uid = key;
                break;
            }
        }
        Log.debug("[getUIDFromLuid] uid=" + uid + " luid="+luid);
        return uid;
    }

   
    protected String getLuidFromUID(String uid) {
        String luid = uid;
        AdditionalItemsAttributes addits   = (AdditionalItemsAttributes)additionalsMap.get(uid);
        if(addits != null){
            luid = addits.getLuid();
        }
        Log.debug("[getLuidFromUID] uid=" + uid + " luid="+luid);
        return luid;
    }

    private PIMItem findItem(String uid) throws PIMException {
        PIMItem item = null;
        if (preloadedItems != null) {
            try {
                item = (PIMItem) preloadedItems.get(uid);
                if (item != null) {
                    Log.debug("Item " + uid + " was preloaded, good!");
                }
            } catch (Exception e) {
                Log.error("Cannot preload item: " + e.toString());
            }
        }

        if (item == null) {
            item = helper.findItem(uid);
        }

        return item;

    }
 
}
