/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.blackberry.ui.rim;

import net.rim.blackberry.api.browser.Browser;
import net.rim.device.api.ui.Font;
import net.rim.device.api.ui.FontFamily;
import net.rim.device.api.ui.component.Dialog;
import net.rim.device.api.ui.container.MainScreen;

import com.funambol.blackberry.SyncClient;
import com.funambol.blackberry.ui.DisplayManager;
import com.funambol.blackberry.ui.Screen;
import com.funambol.util.StaticDataHelper;

public class RimDisplayManager implements DisplayManager {
    
    class AskQuestionRequest implements Runnable {

        private static final int NO_ANSWER = -1;

        private String           message;

        private int              answer    = NO_ANSWER;

        private boolean          answered  = false;

        private boolean          cancelled = false;

        private String[]         options;

        private int[]            values;

        private int              defaultValue;

        private Dialog           dialog    = null;

        private Object           lock      = new Object();

        public AskQuestionRequest(String message, String[] options, int[] values, int defaultValue) {
            this.message = message;
            this.options = options;
            this.values = values;
            this.defaultValue = defaultValue;
            this.answered = false;
            this.lock = new Object();
        }

        public void run() {
            synchronized (SyncClient.getEventLock()) {
                synchronized (lock) {
                    dialog = new Dialog(message, options, values, defaultValue, null);
                }

                dialog.doModal();

                synchronized (lock) {
                    if (!cancelled) {
                        answered = true;
                        answer = dialog.getSelectedValue();
                        lock.notifyAll();
                    }
                    dialog = null;
                }
            }
        }

        public int getAnswer() {
            if (answered) {
                return answer;
            } else {
                return defaultValue;
            }
        }

        public void cancelRequest() {
            class CancelRequest implements Runnable {

                public CancelRequest() {
                }

                public void run() {
                    synchronized (lock) {
                        if (dialog != null) {
                            dialog.cancel();
                            cancelled = true;
                        }
                    }
                }
            }

            SyncClient.getUiApplication().invokeLater(new CancelRequest());
        }

        public int waitForAnswer(long cancelAfterDelay) {
            if (cancelAfterDelay == DisplayManager.NO_LIMIT) {
                cancelAfterDelay = 0;
            }

            try {
                if (SyncClient.getUiApplication().isEventThread()) {
                    this.run();
                } else {
                    synchronized (lock) {
                        SyncClient.getUiApplication().invokeLater(this);
                        lock.wait(cancelAfterDelay);
                    }
                }
            } catch (InterruptedException e) {
                // Do nothing
            }
            synchronized (lock) {
                cancelRequest();

                return getAnswer();
            }
        }

        public int doNotWaitForAnswer(long cancelAfterDelay) {

            synchronized (lock) {
                SyncClient.getUiApplication().invokeLater(this);

                if (cancelAfterDelay != DisplayManager.NO_LIMIT) {
                    final long waitTime = cancelAfterDelay;
                    Thread cancellationThread = new Thread() {

                        public void run() {
                            try {
                                lock.wait(waitTime);
                            } catch (InterruptedException e) {

                            }
                            cancelRequest();
                        }
                    };
                    cancellationThread.start();
                }
            }

            return NO_ANSWER;
        }
    }

    /**
     * Pull a screen to the back
     * 
     * @param scr
     *            The screen to hide
     */
    public void hideScreen(Screen scr) throws Exception {
        if (!(scr instanceof MainScreen)) {
            throw new Exception("The Rim Display Manager can only take MainScreens");
        }

        class HideScreenRequest implements Runnable {

            private MainScreen screen;

            public HideScreenRequest(MainScreen scr) {
                screen = scr;
            }

            public void run() {
                synchronized (SyncClient.getEventLock()) {
                    if (screen.isDisplayed()) {
                        SyncClient.getUiApplication().popScreen(screen);
                    }
                }
            }
        }

        HideScreenRequest request = new HideScreenRequest((MainScreen) scr);

        if (SyncClient.getUiApplication().isEventThread()) {
            request.run();
        } else {
            SyncClient.getUiApplication().invokeLater(request);
        }
    }

    public void showScreen(Screen scr) throws Exception {
        showScreen(scr, false);
    }

    /**
     * Push a screen to the forefront
     * 
     * @param scr
     *            The screen to push
     */
    public void showScreen(Screen scr, boolean force) throws Exception {
        if (!(scr instanceof MainScreen)) {
            throw new Exception("The Rim Display Manager can only take MainScreens");
        }

        class ShowScreenRequest implements Runnable {

            private MainScreen screen;

            public ShowScreenRequest(MainScreen scr) {
                screen = scr;
            }

            public void run() {
                synchronized (SyncClient.getEventLock()) {
                    if (screen.isDisplayed()) {
                        SyncClient.getUiApplication().popScreen(screen);
                    }
                    SyncClient.getUiApplication().pushScreen(screen);
                }
            }
        }

        ShowScreenRequest request = new ShowScreenRequest((MainScreen) scr);

        if (SyncClient.getUiApplication().isEventThread() || force) {
            request.run();
        } else {
            SyncClient.getUiApplication().invokeLater(request);
        }
    }

    /**
     * Helper function to prompt the user for a yes/no answer
     * 
     * @param question
     * @param defaultyes
     * @return True on yes
     */
    public boolean askYesNoQuestion(String question, boolean defaultyes, long timeToWait) {

        String[] buttons = new String[] { StaticDataHelper.getLanguage("dialog_yes"),
                StaticDataHelper.getLanguage("dialog_no") };
        int[] values = new int[] { 1, 0 };

        return 1 == promptSelection(question, buttons, values, (defaultyes ? 1 : 0), true,
                timeToWait);
    }

    /**
     * Helper function to prompt the user for a yes/no answer
     * 
     * @param question
     * @param defaultyes
     * @return True on yes
     */
    public boolean askAcceptDenyQuestion(String question, boolean defaultAccept,
                                         long timeToWait) {

        String[] buttons = new String[] { StaticDataHelper.getLanguage("dialog_accept"),
                StaticDataHelper.getLanguage("dialog_deny") };
        int[] values = new int[] { 1, 0 };

        return 1 == promptSelection(question, buttons, values, (defaultAccept ? 1 : 0), true,
                timeToWait);
    }


    /**
     * Helper function to prompt to continue or cancel
     * 
     * @param message
     * @return True for next
     */
    public boolean promptNext(String message) {
        String[] buttons = new String[] { StaticDataHelper.getLanguage("dialog_next"), StaticDataHelper.getLanguage("dialog_cancel") };
        int[] values = new int[] { 0, 1 };

        return 0 == promptSelection(message, buttons, values, 0, true, DisplayManager.NO_LIMIT);
    }

    /**
     * Helper function to display a simple notice
     * 
     * @param message
     */
    public void showMessage(String message) {
        String[] buttons = new String[] { StaticDataHelper.getLanguage("dialog_ok") };
        int[] values = new int[] { 0 };

        promptSelection(message, buttons, values, 0, false, DisplayManager.NO_LIMIT);
    }

    /**
     * Prompt the user to select an option from a list
     * 
     * @param message
     *            The message to prompt the user with
     * @param options
     *            An array of option names
     * @param values
     *            An array of option values
     * @param defaultValue
     *            The default selected value
     * @param needResponse
     *            TODO
     * @param cancelAfterDelay
     *            TODO
     * @return The chosen value
     */
    public int promptSelection(String message, String[] options, int[] values, int defaultValue,
            boolean needResponse, long cancelAfterDelay) {

        AskQuestionRequest request = new AskQuestionRequest(message, options, values, defaultValue);
        if (needResponse) {
            return request.waitForAnswer(cancelAfterDelay);
        } else {
            return request.doNotWaitForAnswer(cancelAfterDelay);
        }
    }

    public void toForeground() {
        synchronized (SyncClient.getEventLock()) {
            SyncClient.getUiApplication().requestForeground();
        }
    }

    public void toBackground() {

        synchronized (SyncClient.getEventLock()) {
            SyncClient.getUiApplication().requestBackground();
        }
    }

    public void loadBrowser(String url) {
        Browser.getDefaultSession().displayPage(url);
    }

}
